#ifdef VERTEX_SHADER

layout(location = 0) in vec3 a_position;
layout(location = 1) in vec3 a_normal;
layout(location = 3) in vec2 a_uv;

layout (std140, binding = 1) uniform Camera {
    mat4 view;
    mat4 view_inv;
    mat4 projection;
    mat4 projection_inv;
    mat4 viewProjection;
    vec3 camera_position;
    vec4 fov_near_far;
    uvec4 resolution;
    float time;
    vec2 padding;
    float relief_factor;
};

layout(std430, binding=5) buffer Model_in {
	mat4 model_matrix[];
};

out vec3 f_position;
out vec3 f_normal;
out vec2 f_uv;
flat out uint light_index;

uniform sampler2D depth_map;

void main()
{
    float disp_factor = 0.3;
    vec3 displaced = a_position * (1.0 + texture(depth_map, a_uv * (time * 0.05)).x * disp_factor - disp_factor * 0.5 + sin(time * 2) * 0.15);
    f_position = (model_matrix[gl_InstanceID] * vec4(displaced, 1.0)).xyz;
    f_normal = (model_matrix[gl_InstanceID] * vec4(a_normal, 0.0)).xyz;
    f_uv = a_uv;
    light_index = gl_InstanceID;

    gl_Position = viewProjection * vec4(f_position, 1.0);
}

#endif

#ifdef FRAGMENT_SHADER
#define NUM_OF_POINTLIGHTS 3
#define ALPHA_FACTOR 0.4
#define BLUR_FACTOR 1.0

layout (std140, binding = 1) uniform Camera {
    mat4 view;
    mat4 view_inv;
    mat4 projection;
    mat4 projection_inv;
    mat4 viewProjection;
    vec3 camera_position;
    vec4 fov_near_far;
    uvec4 resolution;
    float time;
    vec2 padding;
    float relief_factor;
};

layout (std140, binding = 2) uniform DirectionalLight {
    vec3 direction;
    vec3 color;
} directionalLight;

struct PointLight{
    vec3 position;
    vec3 color;
    vec3 attenuation;
    vec2 near_far_plane;
    vec2 padding;
    vec2 tau_phi;
};

layout (std140, binding = 3) uniform PointLights {
    PointLight pointLights[NUM_OF_POINTLIGHTS];
};

uniform sampler2D diffuse_map;
uniform vec4 materialCoefficients; // x = ambient, y = diffuse, z = specular, w = specular alpha

in vec3 f_position;
in vec3 f_normal;
in vec2 f_uv;
flat in uint light_index;

layout (location = 0) out vec4 FragColor;
layout (location = 1) out vec4 BrightColor;

void main()
{
    vec2 new_uv = (f_uv + time * vec2(0.001)) * (1.0 + sin(time) * 0.1);
    float tex = texture(diffuse_map, new_uv).x;

    FragColor = vec4(tex * pointLights[light_index].color, ALPHA_FACTOR);
    BrightColor = vec4(BLUR_FACTOR * tex * pointLights[light_index].color, 1.0);
}

#endif