#ifdef VERTEX_SHADER

void main()
{
}

#endif

#ifdef GEOMETRY_SHADER

layout (points) in;
layout (triangle_strip, max_vertices = 3) out;
out vec2 uv;

void main(){
    gl_Position = vec4(-1.0, -1.0, 0.0, 1.0);
    uv = vec2(0.0, 0.0);
    EmitVertex();
    gl_Position = vec4(3.0, -1.0, 0.0, 1.0);
    uv = vec2(2.0, 0.0);
    EmitVertex();
    gl_Position = vec4(-1.0, 3.0, 0.0, 1.0);
    uv = vec2(0.0, 2.0);
    EmitVertex();
    EndPrimitive();
}

#endif

#ifdef FRAGMENT_SHADER

#define NUM_OF_POINTLIGHTS 3
#define PI_RCP 0.31830995

in vec2 uv;

layout (std140, binding = 1) uniform Camera {
    mat4 view;
    mat4 view_inv;
    mat4 projection;
    mat4 projection_inv;
    mat4 viewProjection;
    vec3 camera_position;
    vec4 fov_near_far;
    uvec4 resolution;
    float time;
    vec2 padding;
    float relief_factor;
};

struct PointLight{
    vec3 position;
    vec3 color;
    vec3 attenuation;
    vec2 near_far_plane;
    vec2 padding;
    vec2 tau_phi;
};

layout (std140, binding = 3) uniform PointLights {
    PointLight pointLights[NUM_OF_POINTLIGHTS];
};

float dither_pattern[16] = float[16] ( // bayer matrix
0.0f, 0.5f, 0.125f, 0.625f,
0.75f, 0.22f, 0.875f, 0.375f,
0.1875f, 0.6875f, 0.0625f, 0.5625,
0.9375f, 0.4375f, 0.8125f, 0.3125);

uniform sampler2D depth_texture;
uniform samplerCube shadow_map[NUM_OF_POINTLIGHTS];

layout (location = 0) out vec4 FragColor;
layout (location = 1) out vec4 linear_depth;

// from https://stackoverflow.com/questions/32227283/getting-world-position-from-depth-buffer-value
vec3 WorldPosFromDepth(float depth, vec2 uv) {
    float z = depth * 2.0 - 1.0;

    vec4 clipSpacePosition = vec4(uv * 2.0 - 1.0, z, 1.0);
    vec4 viewSpacePosition = projection_inv * clipSpacePosition;

    // Perspective division
    viewSpacePosition /= viewSpacePosition.w;

    vec4 worldSpacePosition = view_inv * viewSpacePosition;

    return worldSpacePosition.xyz;
}

// from learnopengl.com
float linearize_depth(float depth) 
{
    float near = fov_near_far.y;
    float far = fov_near_far.z;
    float z = depth * 2.0 - 1.0; // back to NDC 
    return (2.0 * near * far) / (far + near - z * (far - near));	
}

// shadow map function
float is_lit(vec3 w_position, int i) {
    vec3 light_to_pos = w_position - pointLights[i].position;
    float depth_dist = texture(shadow_map[i], light_to_pos).r * pointLights[i].near_far_plane.y;
    float bias = 0.05;
    return length(light_to_pos) - bias > depth_dist? 0.0 : 1.0;
}


void main()
{
    // Calculate depth value for half resolution image
    vec2 texel_size_full_res = 1.0 / textureSize(depth_texture, 0);
    vec2 texel_uv_offset = texel_size_full_res / 2.0;
    vec2 texel_size_half_res = texel_size_full_res * 2.0;

    vec4 depth_values = vec4(texture(depth_texture, uv + -texel_uv_offset).r,
                             texture(depth_texture, uv + vec2(texel_uv_offset.x, 0.0)).r,
                             texture(depth_texture, uv + vec2(0.0, texel_uv_offset.y)).r,
                             texture(depth_texture, uv + texel_uv_offset).r);

    float max_depth_values = max(max(depth_values.x, depth_values.y), max(depth_values.z, depth_values.w));
    float min_depth_values = min(min(depth_values.x, depth_values.y), min(depth_values.z, depth_values.w));
    float use_max = mod((gl_FragCoord.x + gl_FragCoord.y), 2);
    float resulting_depth_value = use_max * max_depth_values + (1.0 - use_max) * min_depth_values;
    //resulting_depth_value = texture(depth_texture, uv + -texel_uv_offset).r;
    resulting_depth_value = min_depth_values;

    float dither_value = dither_pattern[ (int(gl_FragCoord.x) % 4)*4 + (int(gl_FragCoord.y)%4) ];
      
    float NUM_OF_SAMPLES = 50;
    FragColor = vec4(vec3(0.0),1.0);
    //vec3 w_frag_position = WorldPosFromDepth(texture(depth_texture, uv).r, uv);
    vec3 w_frag_position = WorldPosFromDepth(resulting_depth_value, uv);
    vec3 w_camera_to_frag = w_frag_position - camera_position;
    float raymarch_distance_worldspace =  length(w_camera_to_frag);
    float stepsize = raymarch_distance_worldspace / NUM_OF_SAMPLES;
    vec3 delta_worldspace = w_camera_to_frag / NUM_OF_SAMPLES;

    for(int i = 0; i < NUM_OF_POINTLIGHTS; i++){
        float tau = pointLights[i].tau_phi.x;
        float phi = pointLights[i].tau_phi.y;
        //tau = 0.1;
        //phi = 10.0;
        vec3 sample_pos = w_frag_position + dither_value * delta_worldspace;
        float cont = 0.0;
        for (float l = raymarch_distance_worldspace; l > stepsize; l -= stepsize){
            float lit = is_lit(sample_pos, i);
            float d = length(sample_pos - pointLights[i].position);
            //float att = 1.0 / (pointLights[i].attenuation.x + d * pointLights[i].attenuation.y + d * d * pointLights[i].attenuation.z);
            float att = 1.0 / pow((pointLights[i].attenuation.x + d * pointLights[i].attenuation.y + d * d * pointLights[i].attenuation.z), 2);
            float d_rcp = 1.0 / d;
            // dont cont inside sphere
            if(d < 0.18) att = 0.0;
            //cont += att * tau * (lit * (phi * 0.25 * PI_RCP) * d_rcp * d_rcp) * exp(-d*tau) * exp(-l*tau) * stepsize;
            cont += att * (lit * (phi * 0.25 * PI_RCP) * d_rcp * d_rcp) * exp(-d*tau) * exp(-l*tau) * stepsize;

            sample_pos -= delta_worldspace;
        }
        FragColor += vec4(vec3(pointLights[i].color * cont), 0.0);
    }
    linear_depth = vec4(vec3(linearize_depth(resulting_depth_value)),1.0);
}

#endif

